@@ -10,7 +10,9 @@ from pygmtools.numpy_backend import _hung_kernel
 #     Linear Assignment Problem Solvers     #
 #############################################
 
-def hungarian(s: paddle.Tensor, n1: paddle.Tensor=None, n2: paddle.Tensor=None, nproc: int=1) -> paddle.Tensor:
+def hungarian(s: paddle.Tensor, n1: paddle.Tensor=None, n2: paddle.Tensor=None,
+              unmatch1: paddle.Tensor=None, unmatch2: paddle.Tensor=None,
+              nproc: int=1) -> paddle.Tensor:
     """
     Paddle implementation of Hungarian algorithm
     """
@@ -26,13 +28,21 @@ def hungarian(s: paddle.Tensor, n1: paddle.Tensor=None, n2: paddle.Tensor=None,
         n2 = n2.cpu().numpy()
     else:
         n2 = [None] * batch_num
+    if unmatch1 is not None:
+        unmatch1 = -unmatch1.cpu().numpy()
+    else:
+        unmatch1 = [None] * batch_num
+    if unmatch2 is not None:
+        unmatch2 = -unmatch2.cpu().numpy()
+    else:
+        unmatch2 = [None] * batch_num
 
     if nproc > 1:
         with Pool(processes=nproc) as pool:
-            mapresult = pool.starmap_async(_hung_kernel, zip(perm_mat, n1, n2))
+            mapresult = pool.starmap_async(_hung_kernel, zip(perm_mat, n1, n2, unmatch1, unmatch2))
             perm_mat = np.stack(mapresult.get())
     else:
-        perm_mat = np.stack([_hung_kernel(perm_mat[b], n1[b], n2[b]) for b in range(batch_num)])
+        perm_mat = np.stack([_hung_kernel(perm_mat[b], n1[b], n2[b], unmatch1[b], unmatch2[b]) for b in range(batch_num)])
 
     perm_mat = paddle.to_tensor(perm_mat, place=device)
 
@@ -40,6 +50,7 @@ def hungarian(s: paddle.Tensor, n1: paddle.Tensor=None, n2: paddle.Tensor=None,
 
 
 def sinkhorn(s: paddle.Tensor, nrows: paddle.Tensor=None, ncols: paddle.Tensor=None,
+             unmatchrows: paddle.Tensor=None, unmatchcols: paddle.Tensor=None,
              dummy_row: bool=False, max_iter: int=10, tau: float=1., batched_operation: bool=False) -> paddle.Tensor:
     """
     Paddle implementation of Sinkhorn algorithm
@@ -51,12 +62,13 @@ def sinkhorn(s: paddle.Tensor, nrows: paddle.Tensor=None, ncols: paddle.Tensor=N
     else:
         s = s.transpose((0, 2, 1))
         nrows, ncols = ncols, nrows
+        unmatchrows, unmatchcols = unmatchcols, unmatchrows
         transposed = True
 
     if nrows is None:
-        nrows = paddle.to_tensor([s.shape[1] for _ in range(batch_size)], place=s.place, dtype=paddle.int)
+        nrows = paddle.to_tensor([s.shape[1] for _ in range(batch_size)], place=s.place, dtype=paddle.int64)
     if ncols is None:
-        ncols = paddle.to_tensor([s.shape[2] for _ in range(batch_size)], place=s.place, dtype=paddle.int)
+        ncols = paddle.to_tensor([s.shape[2] for _ in range(batch_size)], place=s.place, dtype=paddle.int64)
 
     # ensure that in each dimension we have nrow < ncol
     transposed_batch = nrows > ncols
@@ -72,56 +84,96 @@ def sinkhorn(s: paddle.Tensor, nrows: paddle.Tensor=None, ncols: paddle.Tensor=N
         nrows = new_nrows
         ncols = new_ncols
 
+        if unmatchrows is not None and unmatchcols is not None:
+            unmatchrows_pad = paddle.concat((
+                unmatchrows,
+                paddle.to_tensor(paddle.full((batch_size, unmatchcols.shape[1] - unmatchrows.shape[1]), -float('inf'), dtype=unmatchrows.dtype), place=unmatchrows.place)),
+            axis=1)
+            new_unmatchrows = paddle.where(transposed_batch.reshape((batch_size, 1)), unmatchcols, unmatchrows_pad)[:, :unmatchrows.shape[1]]
+            new_unmatchcols = paddle.where(transposed_batch.reshape((batch_size, 1)), unmatchrows_pad, unmatchcols)
+            unmatchrows = new_unmatchrows
+            unmatchcols = new_unmatchcols
+
     # operations are performed on log_s
-    s = s / tau
+    log_s = s / tau
+    if unmatchrows is not None and unmatchcols is not None:
+        unmatchrows = unmatchrows / tau
+        unmatchcols = unmatchcols / tau
 
     if dummy_row:
-        assert s.shape[2] >= s.shape[1]
-        dummy_shape = list(s.shape)
-        dummy_shape[1] = s.shape[2] - s.shape[1]
+        assert log_s.shape[2] >= log_s.shape[1]
+        dummy_shape = list(log_s.shape)
+        dummy_shape[1] = log_s.shape[2] - log_s.shape[1]
         ori_nrows = nrows
-        nrows = ncols
-        s = paddle.concat((s, paddle.to_tensor(paddle.full(dummy_shape, -float('inf')), place=s.place)), axis=1)
+        nrows = ncols.clone()
+        log_s = paddle.concat((log_s, paddle.to_tensor(paddle.full(dummy_shape, -float('inf'), dtype=log_s.dtype), place=log_s.place)), axis=1)
+        if unmatchrows is not None:
+            unmatchrows = paddle.concat((unmatchrows, paddle.to_tensor(paddle.full((dummy_shape[0], dummy_shape[1]), -float('inf'), dtype=log_s.dtype), place=log_s.place)), axis=1)
         for b in range(batch_size):
-            s[b, ori_nrows[b]:nrows[b], :ncols[b]] = -100
-            s[b, nrows[b]:, :] = -float('inf')
-            s[b, :, ncols[b]:] = -float('inf')
+            log_s[b, ori_nrows[b]:nrows[b], :ncols[b]] = -100
+
+    # assign the unmatch weights
+    if unmatchrows is not None and unmatchcols is not None:
+        new_log_s = paddle.to_tensor(paddle.full((log_s.shape[0], log_s.shape[1]+1, log_s.shape[2]+1), -float('inf'), dtype=log_s.dtype), place=log_s.place)
+        new_log_s[:, :-1, :-1] = log_s
+        log_s = new_log_s
+        for b in range(batch_size):
+            log_s[b, :nrows[b], ncols[b]] = unmatchrows[b, :nrows[b]]
+            log_s[b, nrows[b], :ncols[b]] = unmatchcols[b, :ncols[b]]
+    row_mask = paddle.zeros((batch_size, log_s.shape[1], 1), dtype=paddle.bool)
+    col_mask = paddle.zeros((batch_size, 1, log_s.shape[2]), dtype=paddle.bool)
+    for b in range(batch_size):
+        row_mask[b, :nrows[b], 0] = 1
+        col_mask[b, 0, :ncols[b]] = 1
+    if unmatchrows is not None and unmatchcols is not None:
+        ncols += 1
+        nrows += 1
 
     if batched_operation:
-        log_s = s
+        for b in range(batch_size):
+            log_s[b, nrows[b]:, :] = -float('inf')
+            log_s[b, :, ncols[b]:] = -float('inf')
 
         for i in range(max_iter):
             if i % 2 == 0:
                 log_sum = paddle.logsumexp(log_s, 2, keepdim=True)
-                log_s = log_s - log_sum
+                log_s = log_s - paddle.where(row_mask, log_sum, paddle.zeros_like(log_sum))
                 nan_indices = paddle.nonzero(paddle.isnan(log_s), True)
-                if nan_indices[0].size > 0:
-                    log_s[nan_indices] = -float('inf')
+                assert nan_indices[0].size == 0
             else:
                 log_sum = paddle.logsumexp(log_s, 1, keepdim=True)
-                log_s = log_s - log_sum
+                log_s = log_s - paddle.where(col_mask, log_sum, paddle.zeros_like(log_sum))
                 nan_indices = paddle.nonzero(paddle.isnan(log_s), True)
-                if nan_indices[0].size > 0:
-                    log_s[nan_indices] = -float('inf')
+                assert nan_indices[0].size == 0
 
         ret_log_s = log_s
     else:
-        ret_log_s = paddle.to_tensor(paddle.full((batch_size, s.shape[1], s.shape[2]), -float('inf')), place=s.place, dtype=s.dtype)
+        ret_log_s = paddle.to_tensor(paddle.full((batch_size, log_s.shape[1], log_s.shape[2]), -float('inf')), place=log_s.place, dtype=log_s.dtype)
 
         for b in range(batch_size):
             row_slice = slice(0, nrows[b])
             col_slice = slice(0, ncols[b])
-            log_s = s[b, row_slice, col_slice]
+            log_s_b = log_s[b, row_slice, col_slice]
+            row_mask_b = row_mask[b, row_slice, :]
+            col_mask_b = col_mask[b, :, col_slice]
 
             for i in range(max_iter):
                 if i % 2 == 0:
-                    log_sum = paddle.logsumexp(log_s, 1, keepdim=True)
-                    log_s = log_s - log_sum
+                    log_sum = paddle.logsumexp(log_s_b, 1, keepdim=True)
+                    log_s_b = log_s_b - paddle.where(row_mask_b, log_sum, paddle.zeros_like(log_sum))
                 else:
-                    log_sum = paddle.logsumexp(log_s, 0, keepdim=True)
-                    log_s = log_s - log_sum
+                    log_sum = paddle.logsumexp(log_s_b, 0, keepdim=True)
+                    log_s_b = log_s_b - paddle.where(col_mask_b, log_sum, paddle.zeros_like(log_sum))
 
-            ret_log_s[b, row_slice, col_slice] = log_s
+            ret_log_s[b, row_slice, col_slice] = log_s_b
+
+    if unmatchrows is not None and unmatchcols is not None:
+        ncols -= 1
+        nrows -= 1
+        for b in range(batch_size):
+            ret_log_s[b, :nrows[b] + 1, ncols[b]] = -float('inf')
+            ret_log_s[b, nrows[b], :ncols[b]] = -float('inf')
+        ret_log_s = ret_log_s[:, :-1, :-1]
 
     if dummy_row:
         if dummy_shape[1] > 0:
@@ -133,7 +185,7 @@ def sinkhorn(s: paddle.Tensor, nrows: paddle.Tensor=None, ncols: paddle.Tensor=N
         s_t = ret_log_s.transpose((0, 2, 1))
         s_t = paddle.concat((
             s_t[:, :ret_log_s.shape[1], :],
-            paddle.to_tensor(paddle.full((batch_size, ret_log_s.shape[1], ret_log_s.shape[2]-ret_log_s.shape[1]), -float('inf')), place=s.place)), axis=2)
+            paddle.to_tensor(paddle.full((batch_size, ret_log_s.shape[1], ret_log_s.shape[2]-ret_log_s.shape[1]), -float('inf')), place=log_s.place)), axis=2)
         ret_log_s = paddle.where(transposed_batch.reshape((batch_size, 1, 1)), s_t, ret_log_s)
 
     if transposed:
@@ -395,7 +447,7 @@ def _aff_mat_from_node_edge_aff(node_aff: paddle.Tensor, edge_aff: paddle.Tensor
         if ne1 is None:
             ne1 = [edge_aff.shape[1]] * batch_size
         if ne2 is None:
-            ne2 = [edge_aff.shape[1]] * batch_size
+            ne2 = [edge_aff.shape[2]] * batch_size
     else:
         device = node_aff.place
         dtype = node_aff.dtype
@@ -425,12 +477,13 @@ def _aff_mat_from_node_edge_aff(node_aff: paddle.Tensor, edge_aff: paddle.Tensor
     return paddle.stack(ks, axis=0)
 
 
-def _check_data_type(input: paddle.Tensor):
+def _check_data_type(input: paddle.Tensor, var_name=None):
     """
     Paddle implementation of _check_data_type
     """
     if type(input) is not paddle.Tensor:
-        raise ValueError(f'Expected Paddle Tensor, but got {type(input)}. Perhaps the wrong backend?')
+        raise ValueError(f'Expected Paddle Tensor{f" for variable {var_name}" if var_name is not None else ""}, '
+                         f'but got {type(input)}. Perhaps the wrong backend?')
 
 
 def _check_shape(input, dim_num):
@@ -463,7 +516,7 @@ def _unsqueeze(input, dim):
 
 def _transpose(input, dim1, dim2):
     """
-    Paddle implementaiton of _transpose
+    Paddle implementation of _transpose
     """
     return paddle.transpose(input, (dim2, dim1))
 
