@@ -1,19 +1,23 @@
 import itertools
+import functools
 import torch
 import numpy as np
 from multiprocessing import Pool
 from torch import Tensor
 
 import pygmtools.utils
-from pygmtools.numpy_backend import _hung_kernel
 
 
 #############################################
 #     Linear Assignment Problem Solvers     #
 #############################################
 
+from pygmtools.numpy_backend import _hung_kernel
 
-def hungarian(s: Tensor, n1: Tensor=None, n2: Tensor=None, nproc: int=1) -> Tensor:
+
+def hungarian(s: Tensor, n1: Tensor=None, n2: Tensor=None,
+              unmatch1: Tensor=None, unmatch2: Tensor=None,
+              nproc: int=1) -> Tensor:
     """
     Pytorch implementation of Hungarian algorithm
     """
@@ -29,13 +33,21 @@ def hungarian(s: Tensor, n1: Tensor=None, n2: Tensor=None, nproc: int=1) -> Tens
         n2 = n2.cpu().numpy()
     else:
         n2 = [None] * batch_num
+    if unmatch1 is not None:
+        unmatch1 = -unmatch1.cpu().numpy()
+    else:
+        unmatch1 = [None] * batch_num
+    if unmatch2 is not None:
+        unmatch2 = -unmatch2.cpu().numpy()
+    else:
+        unmatch2 = [None] * batch_num
 
     if nproc > 1:
         with Pool(processes=nproc) as pool:
-            mapresult = pool.starmap_async(_hung_kernel, zip(perm_mat, n1, n2))
+            mapresult = pool.starmap_async(_hung_kernel, zip(perm_mat, n1, n2, unmatch1, unmatch2))
             perm_mat = np.stack(mapresult.get())
     else:
-        perm_mat = np.stack([_hung_kernel(perm_mat[b], n1[b], n2[b]) for b in range(batch_num)])
+        perm_mat = np.stack([_hung_kernel(perm_mat[b], n1[b], n2[b], unmatch1[b], unmatch2[b]) for b in range(batch_num)])
 
     perm_mat = torch.from_numpy(perm_mat).to(device)
 
@@ -43,6 +55,7 @@ def hungarian(s: Tensor, n1: Tensor=None, n2: Tensor=None, nproc: int=1) -> Tens
 
 
 def sinkhorn(s: Tensor, nrows: Tensor=None, ncols: Tensor=None,
+             unmatchrows: Tensor=None, unmatchcols: Tensor=None,
              dummy_row: bool=False, max_iter: int=10, tau: float=1., batched_operation: bool=False) -> Tensor:
     """
     Pytorch implementation of Sinkhorn algorithm
@@ -54,6 +67,7 @@ def sinkhorn(s: Tensor, nrows: Tensor=None, ncols: Tensor=None,
     else:
         s = s.transpose(1, 2)
         nrows, ncols = ncols, nrows
+        unmatchrows, unmatchcols = unmatchcols, unmatchrows
         transposed = True
 
     if nrows is None:
@@ -75,52 +89,94 @@ def sinkhorn(s: Tensor, nrows: Tensor=None, ncols: Tensor=None,
         nrows = new_nrows
         ncols = new_ncols
 
+        if unmatchrows is not None and unmatchcols is not None:
+            unmatchrows_pad = torch.cat((
+                unmatchrows,
+                torch.full((batch_size, unmatchcols.shape[1] - unmatchrows.shape[1]), -float('inf'), device=s.device)),
+            dim=1)
+            new_unmatchrows = torch.where(transposed_batch.view(batch_size, 1), unmatchcols, unmatchrows_pad)[:, :unmatchrows.shape[1]]
+            new_unmatchcols = torch.where(transposed_batch.view(batch_size, 1), unmatchrows_pad, unmatchcols)
+            unmatchrows = new_unmatchrows
+            unmatchcols = new_unmatchcols
+
     # operations are performed on log_s
-    s = s / tau
+    log_s = s / tau
+    if unmatchrows is not None and unmatchcols is not None:
+        unmatchrows = unmatchrows / tau
+        unmatchcols = unmatchcols / tau
 
     if dummy_row:
-        assert s.shape[2] >= s.shape[1]
-        dummy_shape = list(s.shape)
-        dummy_shape[1] = s.shape[2] - s.shape[1]
+        assert log_s.shape[2] >= log_s.shape[1]
+        dummy_shape = list(log_s.shape)
+        dummy_shape[1] = log_s.shape[2] - log_s.shape[1]
         ori_nrows = nrows
-        nrows = ncols
-        s = torch.cat((s, torch.full(dummy_shape, -float('inf')).to(s.device)), dim=1)
+        nrows = ncols.clone()
+        log_s = torch.cat((log_s, torch.full(dummy_shape, -float('inf'), device=log_s.device, dtype=log_s.dtype)), dim=1)
+        if unmatchrows is not None:
+            unmatchrows = torch.cat((unmatchrows, torch.full((dummy_shape[0], dummy_shape[1]), -float('inf'), device=log_s.device, dtype=log_s.dtype)), dim=1)
         for b in range(batch_size):
-            s[b, ori_nrows[b]:nrows[b], :ncols[b]] = -100
-            s[b, nrows[b]:, :] = -float('inf')
-            s[b, :, ncols[b]:] = -float('inf')
+            log_s[b, ori_nrows[b]:nrows[b], :ncols[b]] = -100
+
+    # assign the unmatch weights
+    if unmatchrows is not None and unmatchcols is not None:
+        new_log_s = torch.full((log_s.shape[0], log_s.shape[1]+1, log_s.shape[2]+1), -float('inf'), device=log_s.device, dtype=log_s.dtype)
+        new_log_s[:, :-1, :-1] = log_s
+        log_s = new_log_s
+        for b in range(batch_size):
+            log_s[b, :nrows[b], ncols[b]] = unmatchrows[b, :nrows[b]]
+            log_s[b, nrows[b], :ncols[b]] = unmatchcols[b, :ncols[b]]
+    row_mask = torch.zeros(batch_size, log_s.shape[1], 1, dtype=torch.bool, device=log_s.device)
+    col_mask = torch.zeros(batch_size, 1, log_s.shape[2], dtype=torch.bool, device=log_s.device)
+    for b in range(batch_size):
+        row_mask[b, :nrows[b], 0] = 1
+        col_mask[b, 0, :ncols[b]] = 1
+    if unmatchrows is not None and unmatchcols is not None:
+        ncols += 1
+        nrows += 1
 
     if batched_operation:
-        log_s = s
+        for b in range(batch_size):
+            log_s[b, nrows[b]:, :] = -float('inf')
+            log_s[b, :, ncols[b]:] = -float('inf')
 
         for i in range(max_iter):
             if i % 2 == 0:
                 log_sum = torch.logsumexp(log_s, 2, keepdim=True)
-                log_s = log_s - log_sum
-                log_s[torch.isnan(log_s)] = -float('inf')
+                log_s = log_s - torch.where(row_mask, log_sum, torch.zeros_like(log_sum))
+                assert not torch.any(torch.isnan(log_s))
             else:
                 log_sum = torch.logsumexp(log_s, 1, keepdim=True)
-                log_s = log_s - log_sum
-                log_s[torch.isnan(log_s)] = -float('inf')
+                log_s = log_s - torch.where(col_mask, log_sum, torch.zeros_like(log_sum))
+                assert not torch.any(torch.isnan(log_s))
 
         ret_log_s = log_s
     else:
-        ret_log_s = torch.full((batch_size, s.shape[1], s.shape[2]), -float('inf'), device=s.device, dtype=s.dtype)
+        ret_log_s = torch.full((batch_size, log_s.shape[1], log_s.shape[2]), -float('inf'), device=log_s.device, dtype=log_s.dtype)
 
         for b in range(batch_size):
             row_slice = slice(0, nrows[b])
             col_slice = slice(0, ncols[b])
-            log_s = s[b, row_slice, col_slice]
+            log_s_b = log_s[b, row_slice, col_slice]
+            row_mask_b = row_mask[b, row_slice, :]
+            col_mask_b = col_mask[b, :, col_slice]
 
             for i in range(max_iter):
                 if i % 2 == 0:
-                    log_sum = torch.logsumexp(log_s, 1, keepdim=True)
-                    log_s = log_s - log_sum
+                    log_sum = torch.logsumexp(log_s_b, 1, keepdim=True)
+                    log_s_b = log_s_b - torch.where(row_mask_b, log_sum, torch.zeros_like(log_sum))
                 else:
-                    log_sum = torch.logsumexp(log_s, 0, keepdim=True)
-                    log_s = log_s - log_sum
+                    log_sum = torch.logsumexp(log_s_b, 0, keepdim=True)
+                    log_s_b = log_s_b - torch.where(col_mask_b, log_sum, torch.zeros_like(log_sum))
 
-            ret_log_s[b, row_slice, col_slice] = log_s
+            ret_log_s[b, row_slice, col_slice] = log_s_b
+
+    if unmatchrows is not None and unmatchcols is not None:
+        ncols -= 1
+        nrows -= 1
+        for b in range(batch_size):
+            ret_log_s[b, :nrows[b] + 1, ncols[b]] = -float('inf')
+            ret_log_s[b, nrows[b], :ncols[b]] = -float('inf')
+        ret_log_s = ret_log_s[:, :-1, :-1]
 
     if dummy_row:
         if dummy_shape[1] > 0:
@@ -132,7 +188,7 @@ def sinkhorn(s: Tensor, nrows: Tensor=None, ncols: Tensor=None,
         s_t = ret_log_s.transpose(1, 2)
         s_t = torch.cat((
             s_t[:, :ret_log_s.shape[1], :],
-            torch.full((batch_size, ret_log_s.shape[1], ret_log_s.shape[2]-ret_log_s.shape[1]), -float('inf'), device=s.device)), dim=2)
+            torch.full((batch_size, ret_log_s.shape[1], ret_log_s.shape[2]-ret_log_s.shape[1]), -float('inf'), device=log_s.device)), dim=2)
         ret_log_s = torch.where(transposed_batch.view(batch_size, 1, 1), s_t, ret_log_s)
 
     if transposed:
@@ -274,18 +330,21 @@ def cao_solver(K, X, num_graph, num_node, max_iter, lambda_init, lambda_step, la
     param_lambda = lambda_init
     device = K.device
 
+    def _comp_aff_score(x, k):
+        return pygmtools.utils.compute_affinity_score(x, k, backend='pytorch').unsqueeze(-1).unsqueeze(-1)
+
     for iter in range(max_iter):
         if iter >= iter_boost:
             param_lambda = np.min([param_lambda * lambda_step, lambda_max])
         # pair_con = get_batch_pc_opt(X)
-        pair_aff = pygmtools.utils.compute_affinity_score(X.reshape(-1, n, n), K.reshape(-1, n * n, n * n), backend='pytorch').reshape(m, m)
+        pair_aff = _comp_aff_score(X.reshape(-1, n, n), K.reshape(-1, n * n, n * n)).reshape(m, m)
         pair_aff = pair_aff - torch.eye(m, device=device) * pair_aff
         norm = torch.max(pair_aff)
         for i in range(m):
             for j in range(m):
                 if i >= j:
                     continue
-                aff_ori = pygmtools.utils.compute_affinity_score(X[i, j], K[i, j], backend='pytorch') / norm
+                aff_ori = _comp_aff_score(X[i, j], K[i, j]) / norm
                 con_ori = _get_single_pc_opt(X, i, j)
                 # con_ori = torch.sqrt(pair_con[i, j])
                 if iter < iter_boost:
@@ -295,7 +354,7 @@ def cao_solver(K, X, num_graph, num_node, max_iter, lambda_init, lambda_step, la
                 X_upt = X[i, j]
                 for k in range(m):
                     X_combo = torch.matmul(X[i, k], X[k, j])
-                    aff_combo = pygmtools.utils.compute_affinity_score(X_combo, K[i, j], backend='pytorch') / norm
+                    aff_combo = _comp_aff_score(X_combo, K[i, j]) / norm
                     con_combo = _get_single_pc_opt(X, i, j, X_combo)
                     # con_combo = torch.sqrt(pair_con[i, k] * pair_con[k, j])
                     if iter < iter_boost:
@@ -322,6 +381,9 @@ def cao_fast_solver(K, X, num_graph, num_node, max_iter, lambda_init, lambda_ste
     m, n = num_graph, num_node
     param_lambda = lambda_init
 
+    def _comp_aff_score(x, k):
+        return pygmtools.utils.compute_affinity_score(x, k, backend='pytorch').unsqueeze(-1).unsqueeze(-1)
+
     device = K.device
     mask1 = torch.arange(m).reshape(m, 1).repeat(1, m).to(device)
     mask2 = torch.arange(m).reshape(1, m).repeat(m, 1).to(device)
@@ -332,7 +394,7 @@ def cao_fast_solver(K, X, num_graph, num_node, max_iter, lambda_init, lambda_ste
         if iter >= iter_boost:
             param_lambda = np.min([param_lambda * lambda_step, lambda_max])
 
-        pair_aff = pygmtools.utils.compute_affinity_score(X.reshape(-1, n, n), K.reshape(-1, n * n, n * n), backend='pytorch').reshape(m, m)
+        pair_aff = _comp_aff_score(X.reshape(-1, n, n), K.reshape(-1, n * n, n * n)).reshape(m, m)
         pair_aff = pair_aff - torch.eye(m, device=device) * pair_aff
         norm = torch.max(pair_aff)
 
@@ -340,12 +402,12 @@ def cao_fast_solver(K, X, num_graph, num_node, max_iter, lambda_init, lambda_ste
         X2 = X.reshape(1, m, m, n, n).repeat(m, 1, 1, 1, 1).transpose(1, 2).reshape(-1, n, n)  # X2[i,j,k] = X[k,j]
         X_combo = torch.bmm(X1, X2).reshape(m, m, m, n, n) # X_combo[i,j,k] = X[i, k] * X[k, j]
 
-        aff_ori = (pygmtools.utils.compute_affinity_score(X.reshape(-1, n, n), K.reshape(-1, n * n, n * n)) / norm).reshape(m, m)
+        aff_ori = (_comp_aff_score(X.reshape(-1, n, n), K.reshape(-1, n * n, n * n)) / norm).reshape(m, m)
         pair_con = _get_batch_pc_opt(X)
         con_ori = torch.sqrt(pair_con)
 
         K_repeat = K.reshape(m, m, 1, n * n, n * n).repeat(1, 1, m, 1, 1).reshape(-1, n * n, n * n)
-        aff_combo = (pygmtools.utils.compute_affinity_score(X_combo.reshape(-1, n, n), K_repeat) / norm).reshape(m, m, m)
+        aff_combo = (_comp_aff_score(X_combo.reshape(-1, n, n), K_repeat) / norm).reshape(m, m, m)
         con1 = pair_con.reshape(m, 1, m).repeat(1, m, 1)  # con1[i,j,k] = pair_con[i,k]
         con2 = pair_con.reshape(1, m, m).repeat(m, 1, 1).transpose(1, 2)  # con2[i,j,k] = pair_con[j,k]
         con_combo = torch.sqrt(con1 * con2)
@@ -370,8 +432,11 @@ def mgm_floyd_solver(K, X, num_graph, num_node, param_lambda):
     m, n = num_graph, num_node
     device = K.device
 
+    def _comp_aff_score(x, k):
+        return pygmtools.utils.compute_affinity_score(x, k, backend='pytorch').unsqueeze(-1).unsqueeze(-1)
+
     for k in range(m):
-        pair_aff = pygmtools.utils.compute_affinity_score(X.reshape(-1, n, n), K.reshape(-1, n * n, n * n), backend='pytorch').reshape(m, m)
+        pair_aff = _comp_aff_score(X.reshape(-1, n, n), K.reshape(-1, n * n, n * n)).reshape(m, m)
         pair_aff = pair_aff - torch.eye(m, device=device) * pair_aff
         norm = torch.max(pair_aff)
 
@@ -383,16 +448,16 @@ def mgm_floyd_solver(K, X, num_graph, num_node, param_lambda):
             for j in range(m):
                 if i >= j:
                     continue
-                score_ori = pygmtools.utils.compute_affinity_score(X[i, j], K[i, j], backend='pytorch') / norm
+                score_ori = _comp_aff_score(X[i, j], K[i, j]) / norm
                 X_combo = torch.matmul(X[i, k], X[k, j])
-                score_combo = pygmtools.utils.compute_affinity_score(X_combo, K[i, j], backend='pytorch') / norm
+                score_combo = _comp_aff_score(X_combo, K[i, j]) / norm
 
                 if score_combo > score_ori:
                     X[i, j] = X_combo
                     X[j, i] = X_combo.transpose(0, 1)
 
     for k in range(m):
-        pair_aff = pygmtools.utils.compute_affinity_score(X.reshape(-1, n, n), K.reshape(-1, n * n, n * n), backend='pytorch').reshape(m, m)
+        pair_aff = _comp_aff_score(X.reshape(-1, n, n), K.reshape(-1, n * n, n * n)).reshape(m, m)
         pair_aff = pair_aff - torch.eye(m, device=device) * pair_aff
         norm = torch.max(pair_aff)
 
@@ -401,13 +466,13 @@ def mgm_floyd_solver(K, X, num_graph, num_node, param_lambda):
             for j in range(m):
                 if i >= j:
                     continue
-                aff_ori = pygmtools.utils.compute_affinity_score(X[i, j], K[i, j], backend='pytorch') / norm
+                aff_ori = _comp_aff_score(X[i, j], K[i, j]) / norm
                 con_ori = _get_single_pc_opt(X, i, j)
                 # con_ori = torch.sqrt(pair_con[i, j])
                 score_ori = aff_ori * (1 - param_lambda) + con_ori * param_lambda
 
                 X_combo = torch.matmul(X[i, k], X[k, j])
-                aff_combo = pygmtools.utils.compute_affinity_score(X_combo, K[i, j], backend='pytorch') / norm
+                aff_combo = _comp_aff_score(X_combo, K[i, j]) / norm
                 con_combo = _get_single_pc_opt(X, i, j, X_combo)
                 # con_combo = torch.sqrt(pair_con[i, k] * pair_con[k, j])
                 score_combo = aff_combo * (1 - param_lambda) + con_combo * param_lambda
@@ -422,13 +487,16 @@ def mgm_floyd_fast_solver(K, X, num_graph, num_node, param_lambda):
     m, n = num_graph, num_node
     device = K.device
 
+    def _comp_aff_score(x, k):
+        return pygmtools.utils.compute_affinity_score(x, k, backend='pytorch').unsqueeze(-1).unsqueeze(-1)
+
     mask1 = torch.arange(m).reshape(m, 1).repeat(1, m)
     mask2 = torch.arange(m).reshape(1, m).repeat(m, 1)
     mask = (mask1 < mask2).float().to(device)
     X_mask = mask.reshape(m, m, 1, 1)
 
     for k in range(m):
-        pair_aff = pygmtools.utils.compute_affinity_score(X.reshape(-1, n, n), K.reshape(-1, n * n, n * n), backend='pytorch').reshape(m, m)
+        pair_aff = _comp_aff_score(X.reshape(-1, n, n), K.reshape(-1, n * n, n * n)).reshape(m, m)
         pair_aff = pair_aff - torch.eye(m, device=device) * pair_aff
         norm = torch.max(pair_aff)
 
@@ -440,8 +508,8 @@ def mgm_floyd_fast_solver(K, X, num_graph, num_node, param_lambda):
         X2 = X[k, :].reshape(1, m, n, n).repeat(m, 1, 1, 1).reshape(-1, n, n)  # X[i, j] = X[j, k]
         X_combo = torch.bmm(X1, X2).reshape(m, m, n, n)
 
-        aff_ori = (pygmtools.utils.compute_affinity_score(X.reshape(-1, n, n), K.reshape(-1, n * n, n * n), backend='pytorch') / norm).reshape(m, m)
-        aff_combo = (pygmtools.utils.compute_affinity_score(X_combo.reshape(-1, n, n), K.reshape(-1, n * n, n * n), backend='pytorch') / norm).reshape(m, m)
+        aff_ori = (_comp_aff_score(X.reshape(-1, n, n), K.reshape(-1, n * n, n * n)) / norm).reshape(m, m)
+        aff_combo = (_comp_aff_score(X_combo.reshape(-1, n, n), K.reshape(-1, n * n, n * n)) / norm).reshape(m, m)
 
         score_ori = aff_ori
         score_combo = aff_combo
@@ -452,7 +520,7 @@ def mgm_floyd_fast_solver(K, X, num_graph, num_node, param_lambda):
         X = X * X_mask + X.transpose(0, 1).transpose(2, 3) * (1 - X_mask)
 
     for k in range(m):
-        pair_aff = pygmtools.utils.compute_affinity_score(X.reshape(-1, n, n), K.reshape(-1, n * n, n * n), backend='pytorch').reshape(m, m)
+        pair_aff = _comp_aff_score(X.reshape(-1, n, n), K.reshape(-1, n * n, n * n)).reshape(m, m)
         pair_aff = pair_aff - torch.eye(m, device=device) * pair_aff
         norm = torch.max(pair_aff)
 
@@ -462,8 +530,8 @@ def mgm_floyd_fast_solver(K, X, num_graph, num_node, param_lambda):
         X2 = X[k, :].reshape(1, m, n, n).repeat(m, 1, 1, 1).reshape(-1, n, n)  # X[i, j] = X[j, k]
         X_combo = torch.bmm(X1, X2).reshape(m, m, n, n)
 
-        aff_ori = (pygmtools.utils.compute_affinity_score(X.reshape(-1, n, n), K.reshape(-1, n * n, n * n), backend='pytorch') / norm).reshape(m, m)
-        aff_combo = (pygmtools.utils.compute_affinity_score(X_combo.reshape(-1, n, n), K.reshape(-1, n * n, n * n), backend='pytorch') / norm).reshape(m, m)
+        aff_ori = (_comp_aff_score(X.reshape(-1, n, n), K.reshape(-1, n * n, n * n)) / norm).reshape(m, m)
+        aff_combo = (_comp_aff_score(X_combo.reshape(-1, n, n), K.reshape(-1, n * n, n * n)) / norm).reshape(m, m)
 
         con_ori = torch.sqrt(pair_con)
         con1 = pair_con[:, k].reshape(m, 1).repeat(1, m)
@@ -759,6 +827,339 @@ def gamgm_real(
     return U
 
 
+############################################
+#          Neural Network Solvers          #
+############################################
+
+from pygmtools.pytorch_modules import *
+
+
+class PCA_GM_Net(torch.nn.Module):
+    """
+    Pytorch implementation of PCA-GM and IPCA-GM network
+    """
+    def __init__(self, in_channel, hidden_channel, out_channel, num_layers, cross_iter_num=-1):
+        super(PCA_GM_Net, self).__init__()
+        self.gnn_layer = num_layers
+        for i in range(self.gnn_layer):
+            if i == 0:
+                gnn_layer = Siamese_Gconv(in_channel, hidden_channel)
+            elif 0 < i < self.gnn_layer - 1:
+                gnn_layer = Siamese_Gconv(hidden_channel, hidden_channel)
+            else:
+                gnn_layer = Siamese_Gconv(hidden_channel, out_channel)
+                self.add_module('affinity_{}'.format(i), WeightedInnerProdAffinity(out_channel))
+            self.add_module('gnn_layer_{}'.format(i), gnn_layer)
+            if i == self.gnn_layer - 2:  # only the second last layer will have cross-graph module
+                self.add_module('cross_graph_{}'.format(i), torch.nn.Linear(hidden_channel * 2, hidden_channel))
+                if cross_iter_num <= 0:
+                 self.add_module('affinity_{}'.format(i), WeightedInnerProdAffinity(hidden_channel))
+
+
+    def forward(self, feat1, feat2, A1, A2, n1, n2, cross_iter_num, sk_max_iter, sk_tau):
+        _sinkhorn_func = functools.partial(sinkhorn,
+                                           dummy_row=False, max_iter=sk_max_iter, tau=sk_tau, batched_operation=False)
+        emb1, emb2 = feat1, feat2
+        if cross_iter_num <= 0:
+            # Vanilla PCA-GM
+            for i in range(self.gnn_layer):
+                gnn_layer = getattr(self, 'gnn_layer_{}'.format(i))
+                emb1, emb2 = gnn_layer([A1, emb1], [A2, emb2])
+
+                if i == self.gnn_layer - 2:
+                    affinity = getattr(self, 'affinity_{}'.format(i))
+                    s = affinity(emb1, emb2)
+                    s = _sinkhorn_func(s, n1, n2)
+
+                    cross_graph = getattr(self, 'cross_graph_{}'.format(i))
+                    new_emb1 = cross_graph(torch.cat((emb1, torch.bmm(s, emb2)), dim=-1))
+                    new_emb2 = cross_graph(torch.cat((emb2, torch.bmm(s.transpose(1, 2), emb1)), dim=-1))
+                    emb1 = new_emb1
+                    emb2 = new_emb2
+
+            affinity = getattr(self, 'affinity_{}'.format(self.gnn_layer - 1))
+            s = affinity(emb1, emb2)
+            s = _sinkhorn_func(s, n1, n2)
+
+        else:
+            # IPCA-GM
+            for i in range(self.gnn_layer - 1):
+                gnn_layer = getattr(self, 'gnn_layer_{}'.format(i))
+                emb1, emb2 = gnn_layer([A1, emb1], [A2, emb2])
+
+            emb1_0, emb2_0 = emb1, emb2
+            s = torch.zeros(emb1.shape[0], emb1.shape[1], emb2.shape[1], device=emb1.device)
+
+            for x in range(cross_iter_num):
+                # cross-graph convolution in second last layer
+                i = self.gnn_layer - 2
+                cross_graph = getattr(self, 'cross_graph_{}'.format(i))
+                emb1 = cross_graph(torch.cat((emb1_0, torch.bmm(s, emb2_0)), dim=-1))
+                emb2 = cross_graph(torch.cat((emb2_0, torch.bmm(s.transpose(1, 2), emb1_0)), dim=-1))
+
+                # last layer
+                i = self.gnn_layer - 1
+                gnn_layer = getattr(self, 'gnn_layer_{}'.format(i))
+                emb1, emb2 = gnn_layer([A1, emb1], [A2, emb2])
+                affinity = getattr(self, 'affinity_{}'.format(i))
+                s = affinity(emb1, emb2)
+                s = _sinkhorn_func(s, n1, n2)
+
+        return s
+
+
+pca_gm_pretrain_path = {
+    'voc': ('https://drive.google.com/u/0/uc?export=download&confirm=Z-AR&id=1HM7dWlKLF0vV2ABL-Vlqq4qVtN5N_QSz',
+            '05924bffc97c9773fda233317c8169d7'),
+    'willow': ('https://drive.google.com/u/0/uc?export=download&confirm=Z-AR&id=1SCWDbAb_YCGy5fsgHAniaVdWwVrSQtwT',
+               'db4fe01e9ba1911c1e22f034e2087b7a'),
+    'voc-all': ('https://drive.google.com/u/0/uc?export=download&confirm=Z-AR&id=1_O8jChVyxOq-N7nUhxLxPLyNHSxDfukx',
+                '0491f3064e2b841099e5ee12fac6c7a2')
+}
+
+
+def pca_gm(feat1, feat2, A1, A2, n1, n2,
+           in_channel, hidden_channel, out_channel, num_layers, sk_max_iter, sk_tau,
+           network, pretrain):
+    """
+    Pytorch implementation of PCA-GM
+    """
+    if feat1 is None:
+        forward_pass = False
+        device = torch.device('cpu')
+    else:
+        forward_pass = True
+        device = feat1.device
+    if network is None:
+        network = PCA_GM_Net(in_channel, hidden_channel, out_channel, num_layers)
+        network = network.to(device)
+        if pretrain:
+            if pretrain in pca_gm_pretrain_path:
+                url, md5 = pca_gm_pretrain_path[pretrain]
+                filename = pygmtools.utils.download(f'pca_gm_{pretrain}_pytorch.pt', url, md5)
+                _load_model(network, filename, device)
+            else:
+                raise ValueError(f'Unknown pretrain tag. Available tags: {pca_gm_pretrain_path.keys()}')
+
+    if forward_pass:
+        batch_size = feat1.shape[0]
+        if n1 is None:
+            n1 = torch.tensor([feat1.shape[1]] * batch_size)
+        if n2 is None:
+            n2 = torch.tensor([feat2.shape[1]] * batch_size)
+        result = network(feat1, feat2, A1, A2, n1, n2, -1, sk_max_iter, sk_tau)
+    else:
+        result = None
+    return result, network
+
+
+ipca_gm_pretrain_path = {
+    'voc': ('https://drive.google.com/u/0/uc?export=download&confirm=Z-AR&id=11Iok8YYU1ojtzuja2jhn59zpSJKVnz5Y',
+            '572da07231ea436ba174fde332f2ae6c'),
+    'willow': ('https://drive.google.com/u/0/uc?export=download&confirm=Z-AR&id=1s2mFIwBXgISasGyqVlIOSVej44ihH5Ax',
+               'd9febe4f567bf5a93430b42b11ebd302'),
+}
+
+
+def ipca_gm(feat1, feat2, A1, A2, n1, n2,
+           in_channel, hidden_channel, out_channel, num_layers, cross_iter, sk_max_iter, sk_tau,
+           network, pretrain):
+    """
+    Pytorch implementation of IPCA-GM
+    """
+    if feat1 is None:
+        forward_pass = False
+        device = torch.device('cpu')
+    else:
+        forward_pass = True
+        device = feat1.device
+    if network is None:
+        network = PCA_GM_Net(in_channel, hidden_channel, out_channel, num_layers, cross_iter)
+        network = network.to(device)
+        if pretrain:
+            if pretrain in ipca_gm_pretrain_path:
+                url, md5 = ipca_gm_pretrain_path[pretrain]
+                filename = pygmtools.utils.download(f'ipca_gm_{pretrain}_pytorch.pt', url, md5)
+                _load_model(network, filename, device)
+            else:
+                raise ValueError(f'Unknown pretrain tag. Available tags: {ipca_gm_pretrain_path.keys()}')
+    if forward_pass:
+        batch_size = feat1.shape[0]
+        if n1 is None:
+            n1 = torch.tensor([feat1.shape[1]] * batch_size)
+        if n2 is None:
+            n2 = torch.tensor([feat2.shape[1]] * batch_size)
+        result = network(feat1, feat2, A1, A2, n1, n2, cross_iter, sk_max_iter, sk_tau)
+    else:
+        result = None
+    return result, network
+
+
+class CIE_Net(torch.nn.Module):
+    """
+    Pytorch implementation of CIE graph matching network
+    """
+    def __init__(self, in_node_channel, in_edge_channel, hidden_channel, out_channel, num_layers):
+        super(CIE_Net, self).__init__()
+        self.gnn_layer = num_layers
+        for i in range(self.gnn_layer):
+            if i == 0:
+                gnn_layer = Siamese_ChannelIndependentConv(in_node_channel, hidden_channel, in_edge_channel)
+            elif 0 < i < self.gnn_layer - 1:
+                gnn_layer = Siamese_ChannelIndependentConv(hidden_channel, hidden_channel, hidden_channel)
+            else:
+                gnn_layer = Siamese_ChannelIndependentConv(hidden_channel, out_channel, hidden_channel)
+                self.add_module('affinity_{}'.format(i), WeightedInnerProdAffinity(out_channel))
+            self.add_module('gnn_layer_{}'.format(i), gnn_layer)
+            if i == self.gnn_layer - 2:  # only the second last layer will have cross-graph module
+                self.add_module('cross_graph_{}'.format(i), torch.nn.Linear(hidden_channel * 2, hidden_channel))
+                self.add_module('affinity_{}'.format(i), WeightedInnerProdAffinity(hidden_channel))
+
+    def forward(self, feat_node1, feat_node2, A1, A2, feat_edge1, feat_edge2, n1, n2, sk_max_iter, sk_tau):
+        _sinkhorn_func = functools.partial(sinkhorn,
+                                           dummy_row=False, max_iter=sk_max_iter, tau=sk_tau, batched_operation=False)
+        emb1, emb2 = feat_node1, feat_node2
+        emb_edge1, emb_edge2 = feat_edge1, feat_edge2
+        for i in range(self.gnn_layer):
+            gnn_layer = getattr(self, 'gnn_layer_{}'.format(i))
+            # during forward process, the network structure will not change
+            emb1, emb2, emb_edge1, emb_edge2 = gnn_layer([A1, emb1, emb_edge1], [A2, emb2, emb_edge2])
+
+            if i == self.gnn_layer - 2:
+                affinity = getattr(self, 'affinity_{}'.format(i))
+                s = affinity(emb1, emb2)
+                s = _sinkhorn_func(s, n1, n2)
+
+                cross_graph = getattr(self, 'cross_graph_{}'.format(i))
+                new_emb1 = cross_graph(torch.cat((emb1, torch.bmm(s, emb2)), dim=-1))
+                new_emb2 = cross_graph(torch.cat((emb2, torch.bmm(s.transpose(1, 2), emb1)), dim=-1))
+                emb1 = new_emb1
+                emb2 = new_emb2
+
+        affinity = getattr(self, 'affinity_{}'.format(self.gnn_layer - 1))
+        s = affinity(emb1, emb2)
+        s = _sinkhorn_func(s, n1, n2)
+        return s
+
+
+cie_pretrain_path = {
+    'voc': ('https://drive.google.com/u/0/uc?export=download&confirm=Z-AR&id=1AzQVcIExjxZLv9hI8nNvOUnzbxZhuOnW',
+            '187916041d9454aecedfd1d09c197f29'),
+    'willow': ('https://drive.google.com/u/0/uc?export=download&confirm=Z-AR&id=1j_mwuSeLzhLFJ9a2b0ZZa73S6jipuhvM',
+               '47cf8f5176a3d17faed96f30fa14ecf4'),
+}
+
+
+def cie(feat_node1, feat_node2, A1, A2, feat_edge1, feat_edge2, n1, n2,
+        in_node_channel, in_edge_channel, hidden_channel, out_channel, num_layers, sk_max_iter, sk_tau,
+        network, pretrain):
+    """
+    Pytorch implementation of CIE
+    """
+    if feat_node1 is None:
+        forward_pass = False
+        device = torch.device('cpu')
+    else:
+        forward_pass = True
+        device = feat_node1.device
+    if network is None:
+        network = CIE_Net(in_node_channel, in_edge_channel, hidden_channel, out_channel, num_layers)
+        network = network.to(device)
+        if pretrain:
+            if pretrain in cie_pretrain_path:
+                url, md5 = cie_pretrain_path[pretrain]
+                filename = pygmtools.utils.download(f'cie_{pretrain}_pytorch.pt', url, md5)
+                _load_model(network, filename, device)
+            else:
+                raise ValueError(f'Unknown pretrain tag. Available tags: {cie_pretrain_path.keys()}')
+
+    if forward_pass:
+        batch_size = feat_node1.shape[0]
+        if n1 is None:
+            n1 = torch.tensor([feat_node1.shape[1]] * batch_size)
+        if n2 is None:
+            n2 = torch.tensor([feat_node1.shape[1]] * batch_size)
+        result = network(feat_node1, feat_node2, A1, A2, feat_edge1, feat_edge2, n1, n2, sk_max_iter, sk_tau)
+    else:
+        result = None
+    return result, network
+
+
+class NGM_Net(torch.nn.Module):
+    """
+    Pytorch implementation of NGM network
+    """
+    def __init__(self, gnn_channels, sk_emb):
+        super(NGM_Net, self).__init__()
+        self.gnn_layer = len(gnn_channels)
+        for i in range(self.gnn_layer):
+            if i == 0:
+                gnn_layer = NGMConvLayer(1, 1,
+                                         gnn_channels[i] + sk_emb, gnn_channels[i],
+                                         sk_channel=sk_emb, edge_emb=False)
+            else:
+                gnn_layer = NGMConvLayer(gnn_channels[i - 1] + sk_emb, gnn_channels[i - 1],
+                                         gnn_channels[i] + sk_emb, gnn_channels[i],
+                                         sk_channel=sk_emb, edge_emb=False)
+            self.add_module('gnn_layer_{}'.format(i), gnn_layer)
+        self.classifier = nn.Linear(gnn_channels[-1] + sk_emb, 1)
+
+    def forward(self, K, n1, n2, n1max, n2max, v0, sk_max_iter, sk_tau):
+        _sinkhorn_func = functools.partial(sinkhorn,
+                                           dummy_row=False, max_iter=sk_max_iter, tau=sk_tau, batched_operation=False)
+        emb = v0
+        A = (K != 0).to(K.dtype)
+        emb_K = K.unsqueeze(-1)
+
+        # NGM qap solver
+        for i in range(self.gnn_layer):
+            gnn_layer = getattr(self, 'gnn_layer_{}'.format(i))
+            emb_K, emb = gnn_layer(A, emb_K, emb, n1, n2, sk_func=_sinkhorn_func)
+
+        v = self.classifier(emb)
+        s = v.view(v.shape[0], n2max, -1).transpose(1, 2)
+
+        return _sinkhorn_func(s, n1, n2, dummy_row=True)
+
+
+ngm_pretrain_path = {
+    'voc': ('https://drive.google.com/u/0/uc?export=download&confirm=Z-AR&id=1POvy6J-9UDNy93qJCKu-czh2FCYkykMK',
+            '60dbc7cc882fd88de4fc9596b7fb0f4a'),
+    'willow': ('https://drive.google.com/u/0/uc?export=download&confirm=Z-AR&id=1ZdlUxyeNoIjA74QTr5wxwQ-vBrr2MBaL',
+               'dd13498bb385df07ac8530da87b14cd6'),
+}
+
+
+def ngm(K, n1, n2, n1max, n2max, x0, gnn_channels, sk_emb, sk_max_iter, sk_tau, network, return_network, pretrain):
+    """
+    Pytorch implementation of NGM
+    """
+    if K is None:
+        forward_pass = False
+        device = torch.device('cpu')
+    else:
+        forward_pass = True
+        device = K.device
+    if network is None:
+        network = NGM_Net(gnn_channels, sk_emb)
+        network = network.to(device)
+        if pretrain:
+            if pretrain in ngm_pretrain_path:
+                url, md5 = ngm_pretrain_path[pretrain]
+                filename = pygmtools.utils.download(f'ngm_{pretrain}_pytorch.pt', url, md5)
+                _load_model(network, filename, device)
+            else:
+                raise ValueError(f'Unknown pretrain tag. Available tags: {ngm_pretrain_path.keys()}')
+
+    if forward_pass:
+        batch_num, n1, n2, n1max, n2max, n1n2, v0 = _check_and_init_gm(K, n1, n2, n1max, n2max, x0)
+        v0 = v0 / torch.mean(v0)
+        result = network(K, n1, n2, n1max, n2max, v0, sk_max_iter, sk_tau)
+    else:
+        result = None
+    return result, network
+
+
 #############################################
 #              Utils Functions              #
 #############################################
@@ -812,6 +1213,7 @@ def build_batch(input, return_ori_dim=False):
     else:
         return torch.stack(padded_ts, dim=0)
 
+
 def dense_to_sparse(dense_adj):
     """
     Pytorch implementation of converting a dense adjacency matrix to a sparse matrix
@@ -830,7 +1232,7 @@ def compute_affinity_score(X, K):
     b, n, _ = X.size()
     vx = X.transpose(1, 2).reshape(b, -1, 1)  # (b, n*n, 1)
     vxt = vx.transpose(1, 2)  # (b, 1, n*n)
-    affinity = torch.bmm(torch.bmm(vxt, K), vx)
+    affinity = torch.bmm(torch.bmm(vxt, K), vx).squeeze(-1).squeeze(-1)
     return affinity
 
 
@@ -927,7 +1329,7 @@ def _aff_mat_from_node_edge_aff(node_aff: Tensor, edge_aff: Tensor, connectivity
         if ne1 is None:
             ne1 = [edge_aff.shape[1]] * batch_size
         if ne2 is None:
-            ne2 = [edge_aff.shape[1]] * batch_size
+            ne2 = [edge_aff.shape[2]] * batch_size
     else:
         device = node_aff.device
         dtype = node_aff.dtype
@@ -959,12 +1361,13 @@ def _aff_mat_from_node_edge_aff(node_aff: Tensor, edge_aff: Tensor, connectivity
     return torch.stack(ks, dim=0)
 
 
-def _check_data_type(input: Tensor):
+def _check_data_type(input: Tensor, var_name=None):
     """
     Pytorch implementation of _check_data_type
     """
     if type(input) is not Tensor:
-        raise ValueError(f'Expected Pytorch Tensor, but got {type(input)}. Perhaps the wrong backend?')
+        raise ValueError(f'Expected Pytorch Tensor{f" for variable {var_name}" if var_name is not None else ""}, '
+                         f'but got {type(input)}. Perhaps the wrong backend?')
 
 
 def _check_shape(input, dim_num):
@@ -1007,3 +1410,30 @@ def _mm(input1, input2):
     Pytorch implementation of _mm
     """
     return torch.mm(input1, input2)
+
+
+def _save_model(model, path):
+    """
+    Save PyTorch model to a given path
+    """
+    if isinstance(model, torch.nn.DataParallel):
+        model = model.module
+
+    torch.save(model.state_dict(), path)
+
+
+def _load_model(model, path, device, strict=True):
+    """
+    Load PyTorch model from a given path. strict=True means all keys must be matched
+    """
+    if isinstance(model, torch.nn.DataParallel):
+        module = model.module
+    else:
+        module = model
+    missing_keys, unexpected_keys = module.load_state_dict(torch.load(path, map_location=device), strict=strict)
+    if len(unexpected_keys) > 0:
+        print('Warning: Unexpected key(s) in state_dict: {}. '.format(
+            ', '.join('"{}"'.format(k) for k in unexpected_keys)))
+    if len(missing_keys) > 0:
+        print('Warning: Missing key(s) in state_dict: {}. '.format(
+            ', '.join('"{}"'.format(k) for k in missing_keys)))
